#!/usr/bin/perl

BEGIN
{
    ($dirprg, $main::SCRIPT) = $0 =~ /(.+)\/(.+)/;
    unshift(@INC, "$dirprg/../lib", "$dirprg/../../lib/");

}

use strict;
use ParamParser;
use Lucene;
use iANT::Entry;
use iANT::Cfg;
require "iANT/Utils.pl";

use MIME::Base64;
use File::Temp;

my $O_PARAM;
my @A_OUTPUT_FIELDS = ();
my @A_INDEX_FIELDS  = ();

MAIN:
{

    my $create_index = 0;

    $O_PARAM = New ParamParser(
                               'GETOPTLONG',
                               (
                                "parser_cfg=s",         "store_dir=s",
                                "query=s",              "output_fields=s",
                                "no_output_field_name", "outfile=s",
                                "escape_special_characters"
                                 )
                                 );
    $O_PARAM->SetUsage(my $usage = sub {&Usage();});

    $O_PARAM->AssertFileExists('parser_cfg');
    my $parser_cfg = $O_PARAM->Get('parser_cfg');
    $O_PARAM->Update($parser_cfg, 'A');
    my ($ra_fields_to_parse, $rh_docid) = &GetLuceneFields4Xml($O_PARAM);
    &GetOutputFields($ra_fields_to_parse);
    &GetIndexFields($ra_fields_to_parse);

    $O_PARAM->SetUsage(my $usage = sub {&Usage();});

    $O_PARAM->SetUnlessDefined('output_fields', $O_PARAM->Get('default_out'));

    $O_PARAM->AssertDefined('store_dir');
    $O_PARAM->AssertDefined('query');

    my $analyzer_mod = $O_PARAM->Get('analyzer');
    my $analyzer     = new $analyzer_mod;

    $O_PARAM->SetBehaviour('use_substitution_table');
    $O_PARAM = &iANT::Cfg::Load($O_PARAM);

    my $unix_find = $O_PARAM->Get('unix_find');
    my $unix_gzip = $O_PARAM->Get('unix_gzip');
    my $unix_cat  = $O_PARAM->Get('unix_cat');

    my $store_dir = $O_PARAM->Get('store_dir');
    my $store = Lucene::Store::FSDirectory->getDirectory($store_dir, 0);

    #$doc serait une entree ?

    my $searcher = new Lucene::Search::IndexSearcher($store);
    my $parser = new Lucene::QueryParser($O_PARAM->Get('default_search'), $analyzer);

    my $query_string = $O_PARAM->Get('query');

    $query_string = &EscapeLuceneSpecialCharaters($query_string, 'no')
      if ($O_PARAM->IsDefined('escape_special_characters'));

    #
    # Ici Test si query contient des tests numeriques ou des conditions &
    #	si oui, traitement special
    #	sinon, traitement classique

    my $query    = $parser->parse($query_string);
    my $hits     = $searcher->search($query);
    my $num_hits = $hits->length();
    print STDERR "INFO:$num_hits hits with query $query_string\n\n";
    my $type;
    my $unix_gzip = $O_PARAM->Get('unix_gzip');

    my $output;
    if ($O_PARAM->IsDefined('outfile'))
    {
        my $out = $O_PARAM->Get('outfile');
        $output = new IO::File("+>$out") || Carp::croak("-- result_file -- cannot open the file : $out");
    }
    else
    {
        $output = *STDOUT;
    }

    my @a_output_fields = split(',', $O_PARAM->Get('output_fields'));
    my $xmltag;
    foreach my $output_field (@a_output_fields)
    {
        if (($O_PARAM->IsDefined($output_field . '_xmltag')) && ($O_PARAM->Get($output_field . '_xmltag') ne ''))
        {
            $xmltag = $O_PARAM->Get($output_field . '_xmltag');
            last;
        }
    }
    if (defined $xmltag)
    {
        print $output "<$xmltag>" . "\n";
    }

    for (my $i = 0; $i < $num_hits; $i++)
    {
        my $doc = $hits->doc($i);
        my @a_output_fields = split(',', $O_PARAM->Get('output_fields'));

        foreach my $output_field (@a_output_fields)
        {
            $output_field =~ s/\s//g;
            if (grep (/$output_field/, @A_OUTPUT_FIELDS))
            {
                my $output_field_value = $doc->get($output_field);
                if ($O_PARAM->Get("field:$output_field") =~ /\.file$/i)
                {

                    my $gz_value = MIME::Base64::decode($output_field_value);

                    my $tmp_dir = $O_PARAM->Get('tmp_dir');
                    my ($tmp_fh, $tmp_filename) = File::Temp::tempfile(DIR => $tmp_dir);
                    print $tmp_fh $gz_value;
                    $tmp_fh->close;
                    $output_field_value = `$unix_gzip -cd $tmp_filename`;
                    chomp $output_field_value;
                    if (defined $xmltag)
                    {

                        #<?xml version="1.0"?>
                        $output_field_value =~ s/<\?xml.+\?>\n//;
                    }
                    unlink($tmp_filename);
                }
                my $prefix = '';
                $prefix = "$output_field=" if (!$O_PARAM->IsDefined('no_output_field_name'));
                print $output "$prefix$output_field_value" . "\n";

            }
        }

    }
    if (defined $xmltag)
    {
        print $output "</$xmltag>" . "\n";
    }
    if ($O_PARAM->IsDefined('outfile'))
    {
        $output->close if (defined $output);
    }

}

sub GetOutputFields
{
    my $ra_fields_to_parse = shift;
    foreach my $rh_field_to_parse (@{$ra_fields_to_parse})
    {
        push(@A_OUTPUT_FIELDS, $rh_field_to_parse->{alias}) if ($rh_field_to_parse->{mode} !~ /unstored/i);
    }
    return;
}

sub GetIndexFields
{
    my $ra_fields_to_parse = shift;
    foreach my $rh_field_to_parse (@{$ra_fields_to_parse})
    {
        push(@A_INDEX_FIELDS, $rh_field_to_parse->{alias}) if ($rh_field_to_parse->{mode} !~ /unindexed/i);
    }
    return;
}

sub Usage
{
    my $output_fields = join("\n\t\t\t\t", @A_OUTPUT_FIELDS);
    my $index_fields  = join("\n\t\t\t\t", @A_INDEX_FIELDS);
    my $default_ouput = $O_PARAM->Get('default_out');
    my $default_search = $O_PARAM->Get('default_search');

    print STDOUT <<END;
   
Usage:

   $0 --parser_cfg=<file> --store_dir=<directory> --query=<string> --output_fields=s
   
 Mandatory
    --parser_cfg:   parser configuration file (contains Analyzer name, fields to parse, ...)
    --store_dir :   directory where Lucene indeces are stored
    --query     :   Lucene query string (default search field: $default_search)
                    Search fields:
                                $index_fields
                                
 Optional
 
    --output_fields : comma separated fields to extract
                    default:    $default_ouput
                    values :    $output_fields
    --no_output_field_name : to avoid output_fields display
    --escape_special_characters : to escape special characters in Lucene query
			           list: \ + - ! ( ) { } [ ] ^ " ~ * ?
    --outfile : the output file

END
}
