#!/usr/bin/perl

BEGIN
{
    ($dirprg, $main::SCRIPT) = $0 =~ /(.+)\/(.+)/;
    unshift(@INC, "$dirprg/../lib", "$dirprg/../../lib/");

}

use strict;
use ParamParser;
use Lucene;
use iANT::Cfg;
require "iANT/Utils.pl";

use MIME::Base64;
use File::Temp;

use IO::File;

use Proc::ProcessTable;
use Digest::MD5 qw(md5 md5_hex md5_base64);
use Cwd 'abs_path';
use Carp;
use HTML::TagCloud;

my $O_PARAM;
my @A_OUTPUT_FIELDS = ();
my @A_INDEX_FIELDS  = ();

my %H_DEFAULT = ();
$H_DEFAULT{levels}  = 20;
$H_DEFAULT{terms_number}  = 50;
$H_DEFAULT{color}  = '#CCCCCC';
$H_DEFAULT{mean_factor} = 4;
$H_DEFAULT{word_min_size} = 10;

MAIN:
{

    my $create_index = 0;

    $O_PARAM = New ParamParser(
                               'GETOPTLONG',
                               (
                                "parser_cfg=s",         "store_dir=s",  "query=s", "url=s", "levels=i", "terms_number=i", "color=s", "mean_factor=i", "word_min_size=i", "ignore_list=s"
                                 )
                                 );
    $O_PARAM->SetUsage(my $usage = sub {&Usage();});
    $O_PARAM->AssertDefined('store_dir');
    
    
    my $store_dir_string = $O_PARAM->Get('store_dir');
    if ($store_dir_string =~ /[\*\?]/)
    {
        $store_dir_string = `echo $store_dir_string`;
        chomp $store_dir_string;
    }
    my @a_store_dir = split(/[,\s]/, $store_dir_string);

    foreach my $store_dir (@a_store_dir)
    {
        if (!-d $store_dir)
        {
            print STDERR "ERROR:  $store_dir is not a directory\n";
            next;
        }
        &RemoveWeirdLockFile($store_dir);

        my $store;
        eval {$store = Lucene::Store::FSDirectory->getDirectory($store_dir, 0)};
        Carp::croak("ERROR [Main01]: $@") if ($@);
    }

    $O_PARAM->AssertFileExists('parser_cfg');
    my $parser_cfg = $O_PARAM->Get('parser_cfg');
    $O_PARAM->Update($parser_cfg, 'A');
    my ($ra_fields_to_parse, $rh_docid) = &GetLuceneFields4Xml($O_PARAM);
    &GetOutputFields($ra_fields_to_parse);
    &GetIndexFields($ra_fields_to_parse);

    $O_PARAM->AssertDefined('query');
    $O_PARAM->SetUnlessDefined('levels', $H_DEFAULT{levels});
    $O_PARAM->SetUnlessDefined('terms_number', $H_DEFAULT{terms_number} );
    $O_PARAM->SetUnlessDefined('color', $H_DEFAULT{color});
    $O_PARAM->SetUnlessDefined('mean_factor', $H_DEFAULT{mean_factor});
    $O_PARAM->SetUnlessDefined('word_min_size', $H_DEFAULT{word_min_size});
    
    

    my $analyzer_mod = $O_PARAM->Get('analyzer');
    my $analyzer     = new $analyzer_mod;

    $O_PARAM->SetBehaviour('use_substitution_table');
    $O_PARAM = &iANT::Cfg::Load($O_PARAM);

    
       

    my $query_string = $O_PARAM->Get('query');

    #
    # Ici Test si query contient des tests numeriques ou des conditions &
    #	si oui, traitement special
    #	sinon, traitement classique

    
    my $ra_values = [];
    foreach my $store_dir (@a_store_dir)
    {
        if (!-d $store_dir)
        {
            print STDERR "ERROR:  $store_dir is not a directory\n";
            next;
        }

        $ra_values = &ListValues($store_dir, $analyzer, $O_PARAM->Get('query'), $ra_values, $O_PARAM);
        
    }
    
    my $rh_terms = &AnalyseTerms($ra_values,$O_PARAM);
    &PrintTagCloud($rh_terms,$O_PARAM);
    

}

sub PrintTagCloud
{
    my $rh_terms = shift;
    my $o_param = shift;
    
    my $url = $o_param->Get('url');
    my $levels = $o_param->Get('levels');
    my $terms_number = $o_param->Get('terms_number');
    my $color = $o_param->Get('color');
    
    
    my $o_cloud = HTML::TagCloud->new(levels=>$levels,distinguish_adjacent_tags=>1);
    foreach my $term (keys %{$rh_terms})
    {
        $o_cloud->add($term, "$url$term", $rh_terms->{$term});
    }

    my $html_and_css = $o_cloud->html_and_css($terms_number);
    $html_and_css =~ s/text-decoration/color : $color; text-decoration/g;
    print $html_and_css;
    
}

sub AnalyseTerms
{
		my $ra_values = shift();
        my $o_param = shift;
		my $word_separator = '[,;:\|\s\[\]\(\)\.\']+';
		my $word_min_size = $o_param->Get('word_min_size');
        my $mean_factor = $o_param->Get('mean_factor');
        
        my @a_ignore_list = split (',', $o_param->Get('ignore_list'));

		my %h_terms = ();
		my $nb_values = scalar @{$ra_values};
		foreach my $value_string (@{$ra_values})
		{
				chomp $value_string;
				my @a_terms = split(/$word_separator/,$value_string);
                
				foreach my $term (@a_terms)
				{
                    my $keep = 1;
                    foreach my $regexp (@a_ignore_list)
                    {
        	        	if ($term =~ /$regexp/i)
                        {
                            $keep = 0;
                            last;
                        }                        
                    }
                    if ($keep == 1 )
                    {
                        $term = lc $term;
    					if ((length $term >= $word_min_size) && ($term =~ /[abcdefghijklmnopqrstuvwxyz]/i))
	    				{
		    					$h_terms{$term} = 0 unless defined $h_terms{$term};
			    				$h_terms{$term} ++;
				    	}
                    }
				}
		}
		
        my $sum = 0;
        my $nb_terms = 0;
		foreach my $term (sort {$h_terms{$b} <=> $h_terms{$a}} (keys %h_terms))
		{		
#			$h_terms{$term} = $h_terms{$term}  / $nb_values;
            $sum += $h_terms{$term};
            $nb_terms ++;
		}
        die "ERROR - No term found" if ($nb_terms == 0);
        
        my $mean = $sum / $nb_terms;
#        print "MEAN:$mean\n";
        foreach my $term (keys %h_terms)
		{
            delete $h_terms{$term} if ($h_terms{$term} > $mean_factor * $mean);
            delete $h_terms{$term} if ($h_terms{$term} < $mean/$mean_factor);
            
        }
        return \%h_terms;
        

}




sub GetOutputFields
{
    my $ra_fields_to_parse = shift;
    foreach my $rh_field_to_parse (@{$ra_fields_to_parse})
    {
        push(@A_OUTPUT_FIELDS, $rh_field_to_parse->{alias}) if ($rh_field_to_parse->{mode} !~ /unstored/i);
    }
    return;
}

sub GetIndexFields
{
    my $ra_fields_to_parse = shift;
    foreach my $rh_field_to_parse (@{$ra_fields_to_parse})
    {
        push(@A_INDEX_FIELDS, $rh_field_to_parse->{alias}) if ($rh_field_to_parse->{mode} !~ /unindexed/i);
    }
    return;
}

sub ListValues
{
    my ($store_dir, $analyzer, $field, $ra_terms, $o_param) = @_;
    
    if ($o_param->IsDefined('order_by') && ($o_param->Get('order_by') !~ /$field/))
    {
        Carp::croak("ERROR [ListValues00] order_by is only possible for field $field when using --list_values switch");
    }

    my $store;
    eval {$store = Lucene::Store::FSDirectory->getDirectory($store_dir, 0)};
    Carp::croak("ERROR [ListValues01] $@") if ($@);

    my $reader;
    eval {$reader = Lucene::Index::IndexReader->open($store)};
    Carp::croak("ERROR [ListValues02] $@") if ($@);

    my $num_docs;
    eval {$num_docs = $reader->numDocs()};
    Carp::croak("ERROR [ListValues03] $@") if ($@);

    my @a_results = ();
    my $result;
    for (my $i = 0; $i < $num_docs; $i++)
    {
        my $document;
        eval {$document = $reader->document($i)};
        Carp::croak("ERROR [ListValues04] $@") if ($@);

        eval {$result = $document->get($field)};
        Carp::croak("ERROR [ListValues05] $@") if ($@);

        if ($O_PARAM->Get("field:$field") =~ /unindexed\./i)
        {
            $result = &DecodeIfNecessary($result);
        }
        push(@{$ra_terms}, $result);
    }
    return $ra_terms;
}

sub DecodeIfNecessary
{
    my $data = shift;

    return $data if ($data eq '');

    my $unix_base64 = $O_PARAM->Get('unix_base64');
    my $unix_cat    = $O_PARAM->Get('unix_cat');

    my $tmp_dir = $O_PARAM->Get('tmp_dir');
    my ($tmp_fh, $tmp_filename) = File::Temp::tempfile(DIR => $tmp_dir);
    print $tmp_fh $data;
    $tmp_fh->close;

    my $cmd     = "$unix_base64 -d $tmp_filename > $tmp_filename.decode";
    my $not_b64 = `($cmd) 2>&1`;

    chomp $not_b64;

    if ($not_b64 eq '')
    {

        my $b64_decoded = `$unix_cat $tmp_filename.decode`;
        chomp $b64_decoded;
        my $unix_gzip = $O_PARAM->Get('unix_gzip');
        $data = `$unix_gzip -cd $tmp_filename.decode`;
        chomp $data;
    }

    unlink $tmp_filename;
    unlink "$tmp_filename.decode";

    return $data;
}

sub RemoveWeirdLockFile
{

    my $store_dir = shift;

    my $tries = 5;
    $store_dir =~ s/\/$//;
    my $store_dir_md5 = md5_hex(abs_path($store_dir));
    my $try           = 0;
    while (
         ((-e '/tmp/lucene-' . $store_dir_md5 . '-commit.lock') || (-e '/tmp/lucene-' . $store_dir_md5 . '-write.lock'))
         && ($try <= $tries))
    {
        my $o_process_table = new Proc::ProcessTable;
        foreach my $o_process (@{$o_process_table->table})
        {
            if ($o_process->cmndline =~ /$store_dir/ && $o_process->pid != $$)
            {
                $try++;
                sleep 5;
            }
            else
            {
                unlink('/tmp/lucene-' . $store_dir_md5 . '-commit.lock')
                  if (-e '/tmp/lucene-' . $store_dir_md5 . '-commit.lock');
                unlink('/tmp/lucene-' . $store_dir_md5 . '-write.lock')
                  if (-e '/tmp/lucene-' . $store_dir_md5 . '-write.lock');
                print STDERR "[WARNING] a persitant lock file has been removed\n";
                last;
            }
        }
    }

    if (   (-e '/tmp/lucene-' . $store_dir_md5 . '-commit.lock')
        || (-e '/tmp/lucene-' . $store_dir_md5 . '-write.lock'))
    {
        Carp::croak("a persitant lock file has been found and could not be removed:\n\t");
        my $o_process_table = new Proc::ProcessTable;
        foreach my $o_process (@{$o_process_table->table})
        {
            if ($o_process->cmndline =~ /$store_dir/ && $o_process->pid != $$)
            {
                print STDERR "[WARNING] a program is accessing to your index: " . $o_process->cmndline;
            }
        }
    }
}

sub Usage
{
    my $output_fields = join("\n\t\t\t\t", @A_OUTPUT_FIELDS);
    my $index_fields  = join("\n\t\t\t\t", @A_INDEX_FIELDS);
    my $default_ouput = $O_PARAM->Get('default_out');
    my $default_search = $O_PARAM->Get('default_search');

    print STDOUT <<END;
   
Usage:

   $0 --parser_cfg=<file> --store_dir=<directory> --query=<string> --output_fields=s
   
 Mandatory:
 
    --parser_cfg:   parser configuration file (contains Analyzer name, fields to parse, ...)
    --store_dir :   directory where Lucene indeces are stored
    --query     :   Lucene query string (default search field: $default_search)
                    Search fields:
                                $index_fields

Optionnal:
 
    Tems analysis:
    -----------------------
    --word_min_size=<int> : minimum word size to be kept [$H_DEFAULT{word_min_size}]
    --mean_factor=<int>   : ignore words which appear more than X times or less than X times average word frequency [$H_DEFAULT{mean_factor}]
    --ignore_list=<list,of,regexp> : ignore terms matching one regexp
    
    Display:
    -----------------------
    --levels=<int>        : number of different size levels [$H_DEFAULT{levels}]
    --terms_number=<int>  : maximun number of terms to display [$H_DEFAULT{terms_number}]
    --url=<string>        : URL underlying terms
    --color=<string>      : terms color [$H_DEFAULT{color}]


END
}
